
#pragma once

// Genode includes
#include <base/env.h>
#include <base/heap.h>

// protobuf includes
#include <google/protobuf/message.h>
#include <google/protobuf/util/time_util.h>

// gapfruit includes
#include <tee_session/connection.h>
#include <tee/packet_allocator.h>
#include <tee_protobuf/helpers.h>
#include <sign_session/connection.h>

// stdcxx includes
#include <functional>


namespace Tee {

	using Genode::Allocator_avl;
	using Genode::Env;
	using Genode::Heap;
	using Genode::List;
	using Genode::Signal_context_capability;
	using Genode::error;
	using Genode::memset;

	using Packet_alloc_failed = Tee::Session::Tx::Source::Packet_alloc_failed;

	using google::protobuf::MessageLite;
	using google::protobuf::Message;

	using gapfruit::protobuf::tee::CounterResponse;
	using gapfruit::protobuf::tee::ExecuteResponse;
	using gapfruit::protobuf::tee::QuoteResponse;
	using gapfruit::protobuf::tee::SignedResponse;

	using ErrorResponse = gapfruit::protobuf::tee::Result;

	using Exec_callback = std::function<void(Packet_content_ro, size_t)>;

	class Connection_exec_only;
}


class Tee::Connection_exec_only
{
	private:

		using Signal_handler = Genode::Signal_handler<Connection_exec_only>;

		Env&                   _env;
		Heap                   _heap             { _env.ram(), _env.rm() };
		Allocator_avl          _tx_block_alloc   { &_heap };
		Tee::Connection        _connection;
		const char*            _session_label;
		const Exec_callback&   _callback;

		// signal handlers for connection
		Signal_handler         _rx_data          { _env.ep(), *this, &Connection_exec_only::_handle_rx };
		Signal_handler         _tx_data          { _env.ep(), *this, &Connection_exec_only::_handle_tx };

		//Signal_context_capability  _counter_request  { };

		// methods used by the signal handlers
		void _handle_rx();
		void _handle_tx();

	public:

		enum { DEFAULT_BUF_SIZE = Tee::Packet_allocator::DEFAULT_PACKET_SIZE * 128 };

		Connection_exec_only(Genode::Env& env,
		                        const char* label,
		                        const Exec_callback& callback,
		                        size_t tx_buf_size = DEFAULT_BUF_SIZE,
		                        size_t rx_buf_size = DEFAULT_BUF_SIZE) :
		    _env             { env },
		    _connection      { env, &_tx_block_alloc, tx_buf_size, rx_buf_size, label },
		    _session_label   { label },
		    _callback        { callback }
		{
			_connection.rx_channel()->sigh_packet_avail(_rx_data);
			_connection.rx_channel()->sigh_ready_to_ack(_rx_data);
			_connection.tx_channel()->sigh_ready_to_submit(_tx_data);
			_connection.tx_channel()->sigh_ack_avail(_tx_data);
		}

		~Connection_exec_only() = default;

		/**
		 * Sends a protobuf message
		 *
		 * \param msg    The protobuf message to send.
		 * \param sid    The session id. Sinks need to reply using the received
		 *               session id so the source can track sessions.
		 * \param sign   areference to a connection providing the signing service.
		 *
		 * \throw Transmission_error    E.g. when there is no sink to send to.
		 * \throw Serialize_failed      Protobuf message could not be serialized.
		 * \throw Packet_alloc_failed   Could not allocate packet on connection.
		 */
		void send_execute_response(SignedResponse& msg, const Tee::Session_id sid, Sign::Connection& sign);

		/**
		 * send_execute_request
		 *
		 * \param msg  request to send.
		 *
		 * \throw Packet_alloc_failed   Could not allocate packet on connection.
		 */
		void send_execute_request(ExecuteRequest& msg);
};
